# Makefile for the project

# Prefixes
SRCDIR=src
INCDIR=inc
CUBADIR=/afs/cern.ch/user/r/rioutine/Programming/Cuba-4.0

# Compilation flags
CXX=g++
# CXX_EXDIFF=-std=c++11 -O2  -pedantic -W -Wall -Wshadow -fPIC -I$(INCDIR)
# CXX_EXDIFF=-std=c++11 -I$(INCDIR)
CXX_EXDIFF=-std=c++11 -g -O2 -fPIC -I$(INCDIR)

# ExDiff compilation

# linked to HEPMC2 thorough PYTHIA8 ===========================

# HEPMC2.
ExDiffHEPMC2: mainExDiffHEPMC2.cc $(PREFIX_LIB)/libpythia8.a libinterface.a
ifeq ($(HEPMC2_USE),true)
	$(CXX) $< -o $@ -std=c++11 -I$(HEPMC2_INCLUDE) $(CXX_COMMON)\
	-I$(INCDIR) -L. -linterface\
	 -L$(HEPMC2_LIB) -Wl,-rpath,$(HEPMC2_LIB) -lHepMC
else
	@echo "Error: $@ requires HEPMC2"
endif

# linked to HEPMC3 thorough PYTHIA8 ===========================

# HEPMC3.
ExDiffHEPMC3: mainExDiffHEPMC3.cc $(PREFIX_LIB)/libpythia8.a libinterface.a
ifeq ($(HEPMC3_USE),true)
	$(CXX) $< -o $@ -std=c++11 -I$(HEPMC3_INCLUDE) $(CXX_COMMON)\
	-I$(INCDIR) -L. -linterface\
	 -L$(HEPMC3_LIB) -Wl,-rpath,$(HEPMC3_LIB) -lHepMC
else
	@echo "Error: $@ requires HEPMC3"
endif

# linked to ROOT and PYTHIA8 ==================================

ExDiffROOT: mainExDiffROOT.cc $(PREFIX_LIB)/libpythia8.a libinterface.a main92.so
ifeq ($(ROOT_USE),true)
	$(CXX) $< main92.so -o $@ -std=c++11 -w -I$(ROOT_INCLUDE) $(CXX_COMMON)\
	-I$(INCDIR) -L. -linterface\
	 `$(ROOTBIN)root-config --cflags` -Wl,-rpath,./\
	 -Wl,-rpath,$(ROOT_LIB) `$(ROOT_BIN)root-config --glibs`
else
	@echo "Error: $@ requires ROOT"
endif
main92.so: main92Dct.cc $(PREFIX_LIB)/libpythia8.a
	$(CXX) $< -o $@ -w -I$(ROOT_INCLUDE) $(CXX_SHARED) $(CXX_COMMON)\
	 `$(ROOTBIN)root-config --cflags`
main92Dct.cc: $(INCDIR)/main92.h $(INCDIR)/main92LinkDef.h
	export LD_LIBRARY_PATH=$$LD_LIBRARY_PATH:$(ROOT_LIB);\
	 $(ROOT_BIN)rootcint -f $@ -c -I$(PREFIX_INCLUDE) $^

# linked to PYTHIA8 ===========================================

ExDiffPY: mainExDiffPY.cc $(PREFIX_LIB)/libpythia8.a libinterface.a
	$(CXX) $< -o $@ -std=c++11 $(CXX_COMMON) -I$(INCDIR) -L. -linterface

# for cross-sections calculations only

ExDiffCalc: mainPIPI.o libinterfacex.a $(CUBADIR)/libcuba.a
	$(CXX) $(CXX_EXDIFF) -I$(CUBADIR) -o ExDiffCalc mainPIPI.o -L. -linterfacex -L$(CUBADIR) -lcuba -lm

mainPIPI.o: mainPIPI.cc
	$(CXX) $(CXX_EXDIFF) -I$(CUBADIR) -c  mainPIPI.cc	

libinterfacex.a: Particle.o Kinematics.o  XSection.o
	ar cr libinterfacex.a Particle.o Kinematics.o XSection.o
	
XSection.o: src/XSection.cpp
	$(CXX) $(CXX_EXDIFF) -I$(CUBADIR) -c src/XSection.cpp
	
# stand alone compilation =====================================

ExDiff: mainExDiff.o libinterface.a
	$(CXX) $(CXX_EXDIFF) -o ExDiff mainExDiff.o -L. -linterface

mainExDiff.o: mainExDiff.cc
	$(CXX) $(CXX_EXDIFF) -c mainExDiff.cc
	
libinterface.a: newrand.o NDarray.o NDgrid.o iFunc.o Generator.o Particle.o Kinematics.o Event.o Interface.o
	ar cr libinterface.a newrand.o NDarray.o NDgrid.o iFunc.o Generator.o Particle.o Kinematics.o Event.o Interface.o	
	
newrand.o: src/newrand.cpp
	$(CXX) $(CXX_EXDIFF) -c src/newrand.cpp	
	
NDarray.o: src/NDarray.cpp
	$(CXX) $(CXX_EXDIFF) -c src/NDarray.cpp

NDgrid.o: src/NDgrid.cpp
	$(CXX) $(CXX_EXDIFF) -c src/NDgrid.cpp

iFunc.o: src/iFunc.cpp
	$(CXX) $(CXX_EXDIFF) -c src/iFunc.cpp

Generator.o: src/Generator.cpp
	$(CXX) $(CXX_EXDIFF) -c src/Generator.cpp	
	
Particle.o: src/Particle.cpp
	$(CXX) $(CXX_EXDIFF) -c src/Particle.cpp		

Kinematics.o: src/Kinematics.cpp
	$(CXX) $(CXX_EXDIFF) -c src/Kinematics.cpp	
	
Event.o: src/Event.cpp
	$(CXX) $(CXX_EXDIFF) -c src/Event.cpp	

Interface.o: src/Interface.cpp
	$(CXX) $(CXX_EXDIFF) -c src/Interface.cpp	