/* 
 * Interface.h
 *
 *  Created on: 17 NOV 2016
 *      Author: Roman Ryutin
 */

#ifndef INTERFACE_HH 
#define INTERFACE_HH

#include <fstream>
#include <cassert>
#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <stdio.h>
#include <time.h>
#include <iomanip>
#include <limits>
#include <string>
#include <ctime>
 
#include "I.h"
#include "PI.h"
#include "newrand.h"
#include "NDarray.h"
#include "NDgrid.h"
#include "iFunc.h"
#include "Generator.h"
#include "Particle.h"
#include "Constants.h"
#include "Kinematics.h"
#include "Event.h"

namespace ExDiff{

// aux class for interface
class Interface{
  public:

// total c.-s. variable	
    double CSTOT;  
// variables for cuts
    int SETVAR, SETCUT;	

// CONSTRUCTORS ------------	
    Interface():
 IDauthors(1),
 IDprocess(0),
 IDenergy(0),
 version(1),
 kinformat(0),
 Nevents(100),
 outformat(0),
 CSTOT(1.0),
 SETVAR(0),
 SETCUT(0)
{
	
	assert(Nevents > 0);	
	assert(IDauthors > 0);
	assert(IDprocess >= 0);
	assert(IDenergy >= 0);
	assert(version > 0);
	assert(kinformat >= 0);
	assert(outformat >= 0);
	
	std::cout << "Interface constructor Default ---------------------------"  << std::endl; 

}
//		
    Interface(const char* filename_):
 IDauthors(1),
 IDprocess(0),
 IDenergy(0),
 version(1),
 kinformat(0),
 Nevents(100),
 outformat(0),
 CSTOT(1.0),
 SETVAR(0),
 SETCUT(0)
{
 
	assert(Nevents > 0);	
	assert(IDauthors > 0);
	assert(IDprocess >= 0);
	assert(IDenergy >= 0);
	assert(version > 0);
	assert(kinformat >= 0);
	assert(outformat >= 0);

	std::cout << "Interface constructor basic ---------------------------"  << std::endl; 
    GetCard(filename_);		
}
//	
	Interface(Interface& interface_){}
    virtual ~Interface(){}
	
	std::vector<double> MassesInit();
	std::vector<int> IDsInit();
	
// Generator initialization functions
   Generator GeneratorInit();   
   
// Kinematics initialization functions
   KinCM KinematicsInit(Generator&);   
   
// Event initialization functions
   Event EventInit(KinCM&);   		
	
// Access to variables	
// get IDauthors const
    inline const int& _IDauthors() const{
      return IDauthors;
    };	
// get IDauthors
    inline int& _IDauthors(){
      return IDauthors;
    };
	
// get IDprocess const
    inline const int& _IDprocess() const{
      return IDprocess;
    };	
// get IDprocess
    inline int& _IDprocess(){
      return IDprocess;
    };	

// get IDenergy const
    inline const int& _IDenergy() const{
      return IDenergy;
    };	
// get IDenergy
    inline int& _IDenergy(){
      return IDenergy;
    };

// get version const
    inline const int& _version() const{
      return version;
    };	
// get version
    inline int& _version(){
      return version;
    };
	
// get Nevents const
    inline const int& _Nevents() const{
      return Nevents;
    };	
// get Nevents
    inline int& _Nevents(){
      return Nevents;
    };	
	
// get format const
    inline const int& _kinformat() const{
      return kinformat;
    };	
// get format
    inline int& _kinformat(){
      return kinformat;
    };		
	
// get format const
    inline const int& _outformat() const{
      return outformat;
    };	
// get format
    inline int& _outformat(){
      return outformat;
    };			
	
	
// get energy by IDenergy
    inline double Get_Energy(int _IDenergy){
      double res;
	  if(_IDenergy == 0){res = 13000.0;}
	  if(_IDenergy == 1){res = 2760.0;}
	  if(_IDenergy == 2){res = 7000.0;}
	  if(_IDenergy == 3){res = 8000.0;}
	  if(_IDenergy == 4){res = 14000.0;}
          if(_IDenergy == 5){res = 1960.0;}
          if(_IDenergy == 6){res = 200.0;} 
	  return res;
    };	
	
// get authors by IDauthors
    inline std::string Get_Authors(int _IDauthors){
      std::string res;
	  if(_IDauthors == 1){res = "A.A. Godizov";}
	  if(_IDauthors == 2){res = "R.A. Ryutin";}
	  if(_IDauthors == 3){res = "V.A. Petrov, R.A. Ryutin";}
	  return res;
    };	

// get process by IDprocess
    inline std::string Get_Process(int _IDprocess){
      std::string res;
	  if(_IDprocess == 0){res = "p + p -> p + p (elastic scattering)";}
	  if(_IDprocess == 1){res = "p + p -> p + f_0(1710) + p (exclusive meson central production)";}
//to get total c.-s. for ID=2 we have to sum contributions c.-s.[S=0]+2*c.-s.[S=1]+2*c.-s.[S=2], i.e. [20]+2*[21]+2*[22]	  
	  if(_IDprocess == 2){res = "p + p -> p + f_2(1950) + p (exclusive meson central production)";}
	  if(_IDprocess == 20){res = "p + p -> p + f_2(1950) [S=0] + p (exclusive meson central production)";}
	  if(_IDprocess == 21){res = "p + p -> p + f_2(1950) [S=1] + p (exclusive meson central production)";}
	  if(_IDprocess == 22){res = "p + p -> p + f_2(1950) [S=2] + p (exclusive meson central production)";}
	  if(_IDprocess == 3){res = "p + p -> p + pi + pi + p (exclusive di-pion central production)";}	
      if(_IDprocess == 4){res = "p + p -> p + f_0(1500) + p (exclusive meson central production)";} 
//to get total c.-s. for ID=5 we have to sum contributions c.-s.[S=0]+2*c.-s.[S=1]+2*c.-s.[S=2], i.e. [50]+2*[51]+2*[52]	  
	  if(_IDprocess == 5){res = "p + p -> p + f_2(1270) + p (exclusive meson central production)";}
	  if(_IDprocess == 50){res = "p + p -> p + f_2(1270) [S=0] + p (exclusive meson central production)";}
	  if(_IDprocess == 51){res = "p + p -> p + f_2(1270) [S=1] + p (exclusive meson central production)";}
	  if(_IDprocess == 52){res = "p + p -> p + f_2(1270) [S=2] + p (exclusive meson central production)";}	  
//to get total c.-s. for ID=6 we have to sum contributions c.-s.[S=0]+2*c.-s.[S=1]+2*c.-s.[S=2], i.e. [60]+2*[61]+2*[62]	  
	  if(_IDprocess == 6){res = "p + p -> p + f_2(2220) + p (exclusive meson central production)";}
	  if(_IDprocess == 60){res = "p + p -> p + f_2(2220) [S=0] + p (exclusive meson central production)";}
	  if(_IDprocess == 61){res = "p + p -> p + f_2(2220) [S=1] + p (exclusive meson central production)";}
	  if(_IDprocess == 62){res = "p + p -> p + f_2(2220) [S=2] + p (exclusive meson central production)";}
	  return res;
    };		

// print information about version of the generator
    inline void GeneratorInfo(){
       std::cout << "+++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
       std::cout << "########            #####" << std::endl;  
       std::cout << "#                   #    #" << std::endl;
       std::cout << "#                   #     #        ###    ###" << std::endl;
       std::cout << "#                   #      #      #   #  #   #" << std::endl;
       std::cout << "########  #     #   #      #  #   #      #" << std::endl; 
       std::cout << "#          #   #    #      #     ###    ###" << std::endl;
       std::cout << "#           ###     #     #   #   #      #" << std::endl; 
       std::cout << "#          #   #    #    #    #   #      #" << std::endl;
       std::cout << "########  #     #   #####     #   #      #" << std::endl;
       std::cout << "+++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
       std::cout << "     Exclusive         Diffractive PROcesses" << std::endl;
       std::cout << "===============================================" << std::endl;
       std::cout << "================ VERSION 2.0 ==================" << std::endl;
       std::cout << "===============================================" << std::endl;
       std::cout << "==== by Roman Ryutin and Anton Godizov ========" << std::endl;
       std::cout << "==== December 1 2016 ==========================" << std::endl;
       std::cout << "===============================================" << std::endl;
       std::cout << "\n\n" << std::endl;	    
       return;	 
	}	
// print information about process and kinematics, authors, version etc.	
	void PrintPars();
// reading info from card file	
	void GetCard(const char*);
// for generation and reformatting file
	void GenerateFile(Generator&, KinCM&, Event&);

	void GenerateFileCuts(Generator&, KinCM&, Event&);
	
// for generation of single event
    Event GenerateEvent(Generator&, KinCM&, Event&);
	
//	void GenerateFileStandard(int); // or GenerateFile reload !
//	void ReFormatFile(int, const char*, int, const char*);

	
	inline std::string ResultFileName(){
		std::string str3;

        std::stringstream out3;
        out3 << "output/A" << IDauthors << "M" << IDprocess << "E" << IDenergy << "_" << version << "F" << kinformat << ".exdiff";
        str3 = out3.str();
//		std::cout << str <<std::endl;
		return str3;	
	}		

// typical grids
// general grid {x1,x2,...} : {x1min,x2min,...} -> {x1max,x2max,...} step {1/(v[1]-1),1/(v[2]-1),...}; v ={N1,N2, ...};
    inline NDgrid<double> grid_1(const std::vector<size_t>& dim_, const std::vector<double>& xmin_, const std::vector<double>& xmax_){
		
		NDgrid<double> res(dim_);
		
		std::size_t i = 0;
		std::size_t j = 0;
		
//		std::cout << "begin calc grid_1===========================================" << std::endl;
		
        for(i = 0; i < dim_.size(); i++){		
  		  for(j = 0; j < dim_[i]; j++){
		    res.GetVar(i,j) = xmin_[i]+j*(xmax_[i]-xmin_[i])/(dim_[i]-1);	
//            std::cout << res.GetVar(i,j) <<std::endl;			
		  }
		}  

        return res;		  
	} 	
// {t1,t2,phi12,log(xi1)}; t1, t2: 0 -> 1 step 1/128.0, phi12: 0 -> PI step PI/8; log(xi1): lmin -> lmax step (lmax-lmin)/8.0	
    inline NDgrid<double> grid_2(const double& mc_, const double& ximin_, const double& sqs_){	

//		std::cout << "begin calc grid_2===========================" << std::endl;
		
		double lmin = log(ximin_);
		double lmax = log((mc_*mc_+4.0)/ximin_/sqs_/sqs_);
		
		std::vector<double> xmin_ = {0.0,0.0,0.0,lmin};
		std::vector<double> xmax_ = {1.0,1.0,PI,lmax};
		std::vector<size_t> dim_ = {129,129,9,9};
		
	    NDgrid<double> res(dim_); 
		res = grid_1(dim_,xmin_,xmax_);
		
        return res;		
	} 
	
// {Dro,phiro,phi12,M_c,y_c;Eta_j,Phi_j}; Dro: 0 -> 1 step 1/20.0, Dphiro: 0 -> PI/4 step PI/24;
// phi12: 0 -> PI step PI/6; M_c: Mmin=1.01*(mc1+mc2) -> Mmax=3 step (Mcmax-Mcmin)/15.0; y_c : 0 -> ycmax=2 step 0.5;
// Eta_j: 0 -> 1 step 1/7; Phi_j: 0 -> 2*PI step 2*PI/10
// multiply by 8 the total c.-s. calculated from this table, since Dphiro, y_c, Eta_j have simmetric intervals (*2*2*2=*8)	
    inline NDgrid<double> grid_3(const double& mc1_, const double& mc2_, const double& Mcmax_, const double& sqs_){	

//		std::cout << "begin calc grid_3===========================" << std::endl;
		
		double Mcmin_ = 1.54309*(mc1_ + mc2_);
		
		std::vector<double> xmin_ = {0.0,0.0001,0.0,Mcmin_,0.0,0.0,0.0};
		std::vector<double> xmax_ = {1.0,0.25*PI,PI,Mcmax_,2.0,1.0,2*PI};
		std::vector<size_t> dim_ = {21,7,7,16,5,6,9};
		
	    NDgrid<double> res(dim_); 
		res = grid_1(dim_,xmin_,xmax_);
		
        return res;		
	} 	

// {D1,D2,phi12,M_c,V_y_c;V_Eta_j,Phi_j}; 
// D1, D2: 0.001, 0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.6, 0.8, 1.0, 1.2, 1.5;
// phi12: 0 -> PI step PI/5; 
// M_c: Mmin=1.54309*(mc1+mc2), 0.5, 0.6, 0.65, 0.7, 0.75, 0.8, 0.9, 1.0, 1.2, 1.4, 1.7, 2.0, 2.3, 2.6, Mmax=3.0; 
// V_y_c: 0.0, +-0.01, +-0.03, +-0.07, +-0.14, +-0.25, +-0.5, +-1.0;
// V_Eta_j: 0, 0.05, 0.1, 0.2, 0.3, 0.5, 0.7, 1.0; 
// Phi_j: 0 -> 2*PI step 2*PI/5;
// multiply by 2 the total c.-s. calculated from this table, Eta_j have simmetric intervals (*2)	
    inline NDgrid<double> grid_4(const double& mc1_, const double& mc2_){	

//		std::cout << "begin calc grid_4===========================" << std::endl;
		
		double Mcmin_ = 1.54309*(mc1_ + mc2_);
		
		std::vector<size_t> dim_ = {15,15,6,15,15,8,6};
		
		std::vector<double> Dgrid_ = {sqrt(0.001),sqrt(0.01),sqrt(0.02),sqrt(0.04),sqrt(0.6),sqrt(0.08),sqrt(0.1),sqrt(0.2),sqrt(0.3),sqrt(0.4),sqrt(0.6),sqrt(0.8),1.0,sqrt(1.2),sqrt(1.5)};
		std::vector<double> Mgrid_ = {Mcmin_,0.5,0.6,0.65,0.7,0.75,0.8,0.9,1.0,1.2,1.4,1.7,2.0,2.5,3.0};
		std::vector<double> Vygrid_ = {-1.0,-0.5,-0.25,-0.14,-0.07,-0.03,-0.01,0.0,0.01,0.03,0.07,0.14,0.25,0.5,1.0};
		std::vector<double> Vetajgrid_ = {0.0,0.05,0.1,0.2,0.3,0.5,0.7,1.0};
		std::vector<double> fi0grid_ = {0.0,PI/5.0,2.0*PI/5.0,3.0*PI/5.0,4.0*PI/5.0,PI};
		std::vector<double> fijgrid_ = {0.0,0.4*PI,0.8*PI,1.2*PI,1.6*PI,2.0*PI};
		
	    NDgrid<double> res(dim_); 
		
		std::size_t i = 0;
		std::size_t j = 0;
		
//		std::cout << "begin calc grid_4===========================================" << std::endl;
        for(i = 0; i < 2; i++){
  		  for(j = 0; j < dim_[i]; j++){
		    res.GetVar(i,j) = Dgrid_[j];			
		  }		
		}
		
        i=2;
  		  for(j = 0; j < dim_[i]; j++){
		    res.GetVar(i,j) = fi0grid_[j];			
		  }		
		  
        i=3;
  		  for(j = 0; j < dim_[i]; j++){
		    res.GetVar(i,j) = Mgrid_[j];			
		  }		  
		  
        i=4;
  		  for(j = 0; j < dim_[i]; j++){
		    res.GetVar(i,j) = Vygrid_[j];			
		  }		  
		  
        i=5;
  		  for(j = 0; j < dim_[i]; j++){
		    res.GetVar(i,j) = Vetajgrid_[j];			
		  }		  
		  
        i=6;
  		  for(j = 0; j < dim_[i]; j++){
		    res.GetVar(i,j) = fijgrid_[j];			
		  }		  
		
        return res;		
	} 		
	
	
	inline std::string DatFileName(){
		std::string str1;
		
        std::stringstream out1;
        out1 << "modeldata/A" << IDauthors << "M" << IDprocess << "E" << IDenergy << "_" << version << "F" << kinformat << ".dat";
        str1 = out1.str();
		
		return str1;			
	}
	inline std::string GridFileName(){ 
		std::string str2; 

        std::stringstream out2;
        out2 << "modeldata/A" << IDauthors << "M" << IDprocess << "E" << IDenergy << "_" << version << "F" << kinformat << ".grid";
        str2 = out2.str();
		
		return str2;	
	}		

 	   
  private:
    int IDauthors, IDprocess, IDenergy, version, Nevents, kinformat, outformat; 

}; // end of class Interface definitions

}// end of namespace ExDiff

#endif /* INTERFACE_HH */
