/*
 * Kinematics.h
 *
 *  Created on: 17 NOV 2016
 *      Author: Roman Ryutin
 */

#ifndef KINEMATICS_HH
#define KINEMATICS_HH

#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <cassert>

#include "I.h"
#include "PI.h"
#include "Particle.h"
#include "Constants.h"

namespace ExDiff{

// aux class for kinematics definition
class KinCM{
  public:
  
  double froot;
  
    KinCM():
 particles(PInit0()),
 variables(Zero()),
 sqs(13000.0),
 t(0.0), 
 phi(0.0), 
 theta(0.0),
 y(0.0),
 DeltaT(0.0), 
 Delta3D(0.0),
 t_1(0.0), 
 t_2(0.0), 
 phi_12(0.0), 
 phi_1(0.0), 
 xi_1(0.0), 
 xi_2(0.0), 
 y_c(0.0), 
 M_T(0.0), 
 M_c(0.0), 
 DeltaT_1(0.0), 
 DeltaT_2(0.0),
 Delta3D_1(0.0), 
 Delta3D_2(0.0),
 t_hat(0.0), 
 s_hat(0.0), 
 Eta_j(0.0), 
 Theta_j(0.0), 
 Phi_j(0.0),
 y_a(0.0),
 y_b(0.0),
 k_ab_x(0.0),
 k_ab_y(0.0),
 phi_ab(0.0),
 froot(1.0),
 hadron1(Hadron1(13000.0)),
 hadron2(Hadron2(13000.0)),
 Delta(Zero()),
 Delta_1(Zero()),
 Delta_2(Zero()),
 flag(0),
 ID(0),
 masses(Zero()),
 IDs(Zeroint())
{
	assert(!particles.empty());
    assert(!variables.empty());	
	
//	std::cout<< "EMPTY CONSTRUCTOR" <<std::endl;
}
//		
    KinCM(const double& sqs_, const std::vector<Particle>& particles_):
 particles(particles_),
 variables(Zero()),
 sqs(sqs_),
 t(0.0), 
 phi(0.0), 
 theta(0.0),
 y(0.0),
 DeltaT(0.0), 
 Delta3D(0.0),
 t_1(0.0), 
 t_2(0.0), 
 phi_12(0.0), 
 phi_1(0.0), 
 xi_1(0.0), 
 xi_2(0.0), 
 y_c(0.0), 
 M_T(0.0), 
 M_c(0.0), 
 DeltaT_1(0.0), 
 DeltaT_2(0.0), 
 Delta3D_1(0.0), 
 Delta3D_2(0.0),
 t_hat(0.0), 
 s_hat(0.0), 
 Eta_j(0.0), 
 Theta_j(0.0), 
 Phi_j(0.0),
 y_a(0.0),
 y_b(0.0),
 k_ab_x(0.0),
 k_ab_y(0.0),
 phi_ab(0.0), 
 froot(1.0),
 hadron1(Hadron1(sqs_)),
 hadron2(Hadron2(sqs_)),
 Delta(Zero()),
 Delta_1(Zero()),
 Delta_2(Zero()),
 flag(0),
 ID(0),
 masses(Zero()),
 IDs(Zeroint())
{
	assert(!particles.empty());
//	std::cout<< "P to V CONSTRUCTOR" <<std::endl;
	

	if(particles.size() < 2 || particles.size() > 4){
		std::cout<< "It is not possible to define 2 to 2 or 3 or 4 kinematics, number of particles are irrelevant!" <<std::endl;	
		flag = 1;
	} else {	
             int n = Phys(); 	
             if(n == 0){   			 
	           if(particles.size() == 2){		   
	 	         PtoV2();
	           }
	           if(particles.size() == 3 ){
  	           	 PtoV3();
	           }	
	           if(particles.size() == 4){
	 	         PtoV4();
	           }
               
	           masses.clear();
	           IDs.clear();
			   int i = 0;
			   for(i = 0; i < particles.size(); i++){
				  masses.push_back(particles[i]._m());
                  IDs.push_back(particles[i]._ID());				  
			   }
			   
	         } else {
				 flag = 1;
				 std::cout<< "Physical region is impossible!" <<std::endl;
				 if(n == 2 || n == 4){
					std::cout<< "Some particles are off-shell!" <<std::endl; 
				 }	
				 if(n == 3){
					std::cout<< "C.M. subsystem energy is not enough to produce hadrons!" <<std::endl; 
				 }		
				 if(n == 5){
					std::cout<< "Some procedures are irrelevant!" <<std::endl; 
				 }						 
			   }
	  }
	
}
//	
    KinCM(const double& sqs_, const std::size_t& ID_, const std::vector<double>& variables_, const std::vector<double>& masses_, const std::vector<int>& IDs_):
 particles(PInit()),
 variables(variables_),
 sqs(sqs_),
 t(0.0), 
 phi(0.0), 
 theta(0.0),
 y(0.0),
 DeltaT(0.0), 
 Delta3D(0.0),
 t_1(0.0), 
 t_2(0.0), 
 phi_12(0.0), 
 phi_1(0.0), 
 xi_1(0.0), 
 xi_2(0.0), 
 y_c(0.0), 
 M_T(0.0), 
 M_c(0.0), 
 DeltaT_1(0.0), 
 DeltaT_2(0.0), 
 Delta3D_1(0.0), 
 Delta3D_2(0.0),
 t_hat(0.0), 
 s_hat(0.0), 
 Eta_j(0.0), 
 Theta_j(0.0), 
 Phi_j(0.0),
 y_a(0.0),
 y_b(0.0),
 k_ab_x(0.0),
 k_ab_y(0.0),
 phi_ab(0.0), 
 froot(1.0),
 hadron1(Hadron1(sqs_)),
 hadron2(Hadron2(sqs_)),
 Delta(Zero()),
 Delta_1(Zero()),
 Delta_2(Zero()),
 flag(0),
 ID(ID_),
 masses(masses_),
 IDs(IDs_) 
{
	assert(!variables.empty());
//	std::cout<< "V to P CONSTRUCTOR" <<std::endl;
// define particles	
    std::vector<double> masses__ = masses_;
	std::vector<int> IDs__ = IDs_;
	std::size_t ID__ = ID_;
	if(masses__.size() < 2 || masses__.size() > 4){
		flag = 1;
		std::cout<< "It is not possible to define 2 to 2 or 3 or 4 kinematics, number of particles are irrelevant!" <<std::endl;	
	} else {	
	         if(masses__.size() == 2){
	 	      VtoP2(ID__,masses__,IDs__);
	         }
	         if(masses__.size() == 3){
  	  	      VtoP3(ID__,masses__,IDs__);
	         }	
	         if(masses__.size() == 4){
	 	      VtoP4(ID__,masses__,IDs__);
	         }		
	  }

      int n = 0;

      if(flag != 0){
		  n = 5;
		  std::cout<< "Some calculations are irrelevant!" <<std::endl; 
	  }
	  
 	  n = Phys();
	  if(n != 0 || flag != 0){
	    std::cout<< "Physical region is impossible! Check out values of variables!" <<std::endl;	 
				 if(n == 2 || n == 4){
					std::cout<< "Some particles are off-shell!" <<std::endl; 
				 }	
				 if(n == 3){
					std::cout<< "C.M. subsystem energy is not enough to produce hadrons!" <<std::endl; 
				 }	
				 if(n == 5){
					std::cout<< "Some calculations are irrelevant!" <<std::endl; 
				 }					 
	  }
		
}
//	
	KinCM(KinCM& kin){}
    virtual ~KinCM(){}
	
	void Print();
// reset variables in the same type of set -------------------------------------------------------------------------
    inline void ResetVars(const std::vector<double>& variables_){
        
// define particles	
      std::vector<double> masses__ = masses;
	  std::vector<int> IDs__ = IDs;
	  std::size_t ID__ = ID;
	  variables = variables_;
	  
//	  std::cout << "variables Size = " << variables.size() << "  Reset variables = " << variables[0] << " " << variables[1] << " " << variables[2] << " " << variables[3] << " " << variables[4] << std::endl;
//	  std::cout << "particles Size = " << particles.size() << std::endl;
	  
	  flag = 0;
	if(masses__.size() < 2 || masses__.size() > 4){
		flag = 1;
		std::cout << "It is not possible to define 2 to 2 or 3 or 4 kinematics, number of particles are irrelevant!" <<std::endl;	
	} else {	
	         if(masses__.size() == 2){
	 	      VtoP2(ID__,masses__,IDs__);
	         }
	         if(masses__.size() == 3){
//			  std::cout << "call VtoP3!!!!!!!!!!!!!!!!!!!!!!!!!!!" << std::endl;	 
  	  	      VtoP3(ID__,masses__,IDs__);
	         }	
	         if(masses__.size() == 4){
	 	      VtoP4(ID__,masses__,IDs__); 
	         }		
	  }
///////////////////////////////////////////////////!!!!!!!!!!!!!!!!!!!!!!!!!
//	  std::cout << "variables Size = " << variables.size() << "  Reset variables = " << variables[0] << " " << variables[1] << " " << variables[2] << " " << variables[3] << " " << variables[4] << std::endl;
//	  std::cout << "particles Size = " << particles.size() << std::endl;

	  
      int n = 0;

      if(flag != 0){
		  n = 5;
		  std::cout << "Some calculations are irrelevant!" <<std::endl; 
	  }
	  
 	  n = Phys();
	  if(n != 0 || flag != 0){
	    std::cout << "Physical region is impossible! Check out values of variables!" <<std::endl;	 
				 if(n == 2 || n == 4){
					std::cout << "Some particles are off-shell!" <<std::endl; 
				 }	
				 if(n == 3){
					std::cout << "C.M. subsystem energy is not enough to produce hadrons!" <<std::endl; 
				 }	
				 if(n == 5){
					std::cout << "Some calculations are irrelevant!" <<std::endl; 
				 }					 
	  }
	
	  return;
		
	}
//------------------------------------------------------------------------------------------------------------	
// functions to check physical region and all particles on-shell!
    int Phys();
	
	inline const int& _flag() const{
		return flag;
	};
	inline int& _flag(){
		return flag;
	};		

// functions to get variables    	
	inline const double& _sqs() const{
		return sqs;
	};
	inline double& _sqs(){
		return sqs;
	};	   	

	inline const double& _t() const{
		return t;
	};
	inline double& _t(){
		return t;
	};	
	
	inline const double& _phi() const{
		return phi;
	};
	inline double& _phi(){
		return phi;
	};	

	inline const double& _theta() const{
		return theta;
	};
	inline double& _theta(){
		return theta;
	};	
	
	inline const double& _y() const{
		return y;
	};
	inline double& _y(){
		return y;
	};
	
	inline const double& _DeltaT() const{
		return DeltaT;
	};
	inline double& _DeltaT(){
		return DeltaT;
	};

	inline const double& _Delta3D() const{
		return Delta3D;
	};
	inline double& _Delta3D(){
		return Delta3D;
	};

	inline const double& _t_1() const{
		return t_1;
	};
	inline double& _t_1(){
		return t_1;
	};

	inline const double& _t_2() const{
		return t_2;
	};
	inline double& _t_2(){
		return t_2;
	};	

	inline const double& _phi_12() const{
		return phi_12;
	};
	inline double& _phi_12(){
		return phi_12;
	};

	inline const double& _phi_1() const{
		return phi_1;
	};
	inline double& _phi_1(){
		return phi_1;
	};

	inline const double& _xi_1() const{
		return xi_1;
	};
	inline double& _xi_1(){
		return xi_1;
	};

	inline const double& _xi_2() const{
		return xi_2;
	};
	inline double& _xi_2(){
		return xi_2;
	};

	inline const double& _y_c() const{
		return y_c;
	};
	inline double& _y_c(){
		return y_c;
	};

	inline const double& _M_T() const{
		return M_T;
	};
	inline double& _M_T(){
		return M_T;
	};

	inline const double& _M_c() const{
		return M_c;
	};
	inline double& _M_c(){
		return M_c;
	};

	inline const double& _DeltaT_1() const{
		return DeltaT_1;
	};
	inline double& _DeltaT_1(){
		return DeltaT_1;
	};

	inline const double& _DeltaT_2() const{
		return DeltaT_2;
	};
	inline double& _DeltaT_2(){
		return DeltaT_2;
	};

	inline const double& _Delta3D_1() const{
		return Delta3D_1;
	};
	inline double& _Delta3D_1(){
		return Delta3D_1;
	};

	inline const double& _Delta3D_2() const{
		return Delta3D_2;
	};
	inline double& _Delta3D_2(){
		return Delta3D_2;
	};

	inline const double& _t_hat() const{
		return t_hat;
	};
	inline double& _t_hat(){
		return t_hat;
	};
	
	inline const double& _s_hat() const{
		return s_hat;
	};
	inline double& _s_hat(){
		return s_hat;
	};	

	inline const double& _Eta_j() const{
		return Eta_j;
	};
	inline double& _Eta_j(){
		return Eta_j;
	};

	inline const double& _Theta_j() const{
		return Theta_j;
	};
	inline double& _Theta_j(){
		return Theta_j;
	};	

	inline const double& _Phi_j() const{
		return Phi_j;
	};
	inline double& _Phi_j(){
		return Phi_j;
	};
	
//////////////////	
	inline const double& _y_a() const{
		return y_a;
	};
	inline double& _y_a(){
		return y_a;
	};	
	inline const double& _y_b() const{
		return y_b;
	};
	inline double& _y_b(){
		return y_b;
	};	
// momentum kab=kaT-kbT !	
	inline const double& _k_ab_x() const{
		return k_ab_x;
	};
	inline double& _k_ab_x(){
		return k_ab_x;
	};	
	inline const double& _k_ab_y() const{
		return k_ab_y;
	};
	inline double& _k_ab_y(){
		return k_ab_y;
	};	
	inline const double& _phi_ab() const{
		return phi_ab;
	};
	inline double& _phi_ab(){
		return phi_ab;
	};		
//////////////////	
	inline const Particle& _hadron1() const{
		return hadron1;
	};
	inline Particle& _hadron1(){
		return hadron1;
	};	
	
	inline const Particle& _hadron2() const{
		return hadron2;
	};
	inline Particle& _hadron2(){
		return hadron2;
	};	
	
	inline const std::vector<double>& _Delta() const{
		return Delta;
	};
	inline std::vector<double>& _Delta(){
		return Delta;
	};	

	inline const std::vector<double>& _Delta_1() const{
		return Delta_1;
	};
	inline std::vector<double>& _Delta_1(){
		return Delta_1;
	};

	inline const std::vector<double>& _Delta_2() const{
		return Delta_2;
	};
	inline std::vector<double>& _Delta_2(){
		return Delta_2;
	};	
	
	inline const std::vector<Particle>& _particles() const{
		return particles;
	};
	inline std::vector<Particle>& _particles(){
		return particles;
	};		
	
	std::vector<double> Transform_CMpp_to_CMdd(std::vector<double>&);
	std::vector<double> Transform_CMdd_to_CMpp(std::vector<double>&);
	
	std::vector<double> Transform_CMpp_to_CMdd_(std::vector<double>&,std::vector<double>&,std::vector<double>&);
	std::vector<double> Transform_CMdd_to_CMpp_(std::vector<double>&,std::vector<double>&,std::vector<double>&);

    inline std::vector<Particle> EvParticles(){
	   std::vector<Particle> evpart_;
	   evpart_.clear();
	   evpart_.push_back(hadron1);
	   evpart_.push_back(hadron2);
	   for(std::size_t i = 0; i < particles.size(); i++){
	  	 evpart_.push_back(particles[i]); 
	   }	  
       return evpart_;	  
    } 

// special kinematical cuts check
// first argument - variable to histogram, second argument - version of cuts,
// returns 0, if an event is not satisfied the cuts, 1 in other case.	
	int Cuts(int&);
	
  private:
   Particle hadron1, hadron2;
// vector of particles
   std::vector<Particle> particles;
// vector of variables  
   std::vector<double> variables;
// sqrt(s)  
   double sqs;
// 2 to 2 variables
// theta - scattering angle, y - rapidity of 1st final proton
   double t, phi, DeltaT, Delta3D, theta, y;
   std::vector<double> Delta;
// 2 to 3 variables
   double t_1, t_2, phi_12, phi_1, xi_1, xi_2, y_c, M_T, M_c, DeltaT_1, DeltaT_2, Delta3D_1, Delta3D_2;
   std::vector<double> Delta_1, Delta_2;
// 2 to 4 additional variables for subprocess 2 to 2 in C.M. of the subprocess
   double t_hat, s_hat, Eta_j, Theta_j, Phi_j;
// additional variables in pp C.M.
   double  y_a, y_b, k_ab_x, k_ab_y, phi_ab;
   
   int flag;   
   std::size_t ID;
   std::vector<double> masses;
   std::vector<int> IDs;
 
// functions to define all variables from initial data
   void VtoP2(std::size_t&, std::vector<double>&, std::vector<int>&);
   void VtoP3(std::size_t&, std::vector<double>&, std::vector<int>&);
   void VtoP4(std::size_t&, std::vector<double>&, std::vector<int>&);
   
   void PtoV2();
   void PtoV3();
   void PtoV4();
   
   void DeltasInit();   
   
   inline std::vector<Particle> PInit(){
	   Particle p1 = hadron1;
	   Particle p2 = hadron2;
	   std::vector<Particle> res({p1,p2});
	   return res;
   }
   
   inline std::vector<Particle> PInit0(){
	   Particle p1 = Hadron1(13000.0);
	   Particle p2 = Hadron2(13000.0);
	   std::vector<Particle> res({p1,p2});
	   return res;
   }   
   
   inline Particle Hadron1(const double& sqs_){
	  Constants c; 
	  std::vector<double> p_h1({0.5*sqs_,0.0,0.0,sqrt(0.25*sqs_*sqs_-c.m_p*c.m_p)}); 
	  Particle h1(p_h1,c.m_p,2212,2,0,0);
      return h1;	  
   }
   inline Particle Hadron2(const double& sqs_){
	  Constants c; 
	  std::vector<double> p_h2({0.5*sqs_,0.0,0.0,-sqrt(0.25*sqs_*sqs_-c.m_p*c.m_p)}); 	   
	  Particle h2(p_h2,c.m_p,2212,2,0,0);
      return h2;	  
   }
   
   inline std::vector<double> Zero(){
	  std::vector<double> res({0.0,0.0,0.0,0.0});
      return res;	  
   }   
   
   inline std::vector<int> Zeroint(){
	  std::vector<int> res({0,0,0,0});
      return res;	  
   }    
 
 
}; // end of class KinCM definitions


}// end of namespace ExDiff

#endif /* KINEMATICS_HH */
