/*
 * NDarray.h
 *
 *  Created on: 11 окт. 2016 г.
 *      Author: Roman Ryutin
 */

#ifndef NDARRAY_HH
#define NDARRAY_HH


#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <cassert>


namespace ExDiff{

template<typename T>

class NDarray{
  public:

    explicit NDarray(const std::vector<std::size_t>& _dimensions) : 
	   dimensions(_dimensions), 
	   values(computeTotalSize(_dimensions)) 
	{
       assert(!dimensions.empty());
       assert(!values.empty());
    };

//    NDarray(){}	
//	NDarray(NDarray& ndarray_){}	
    virtual ~NDarray(){};

/////////////////////////////////////////////////		
// get const element of the array
    inline const T& get(const std::vector<std::size_t>& indexes) const{
      return values[computeIndex(indexes)];
    };
// get element of the array	
    inline T& get(const std::vector<std::size_t>& indexes){
       return values[computeIndex(indexes)];
    };
/////////////////////////////////////////////////	
// get dimensions const
    inline const std::vector<size_t>& GetDimensions() const{
      return dimensions;
    };	
// get dimensions
    inline std::vector<size_t>& GetDimensions(){
      return dimensions;
    };	
/////////////////////////////////////////////////	
// get values const
    inline const std::vector<T>& GetValues() const{
      return values;
    };	
// get values
    inline std::vector<T>& GetValues(){
      return values;
    };	
/////////////////////////////////////////////////		
// compute Index of the vector values from indexes of the NDarray
    std::size_t computeIndex(const std::vector<std::size_t>& indexes) const{
        assert(indexes.size() == dimensions.size());

        std::size_t index = 0;
        std::size_t mul = 1;
		std::size_t n = dimensions.size();
     
        for (std::size_t i = n; i != 0; --i) {
           assert(indexes[i - 1] < dimensions[i - 1]);
           index += indexes[i - 1] * mul;
           mul *= dimensions[i - 1];
        }
        assert(index < values.size());
        return index;
    };
/////////////////////////////////////////////////		
// compute Indexes of the NDarray from the Index of the vector values
    std::vector<std::size_t> computeIndexes(std::size_t index) const{
        assert(index < values.size());

        std::vector<std::size_t> res(dimensions.size());
     
        std::size_t mul = values.size();
		std::size_t n = dimensions.size();
        for (std::size_t i = 0; i != n; ++i) {
         mul /= dimensions[i];
         res[i] = index / mul;
         assert(res[i] < dimensions[i]);
         index -= res[i] * mul;
        }
        return res;		
		
    };
/////////////////////////////////////////////////		
// print information about NDarray
    inline void printC(){
     std::cout<<"NDarray OBJECT INFORMATION ========================= " <<std::endl;
	 std::cout<<"NDarray dimensions SIZE = "<< dimensions.size() <<std::endl;
	 std::cout<<"NDarray dimensions vector" <<std::endl;
	 PrintVector(dimensions,dimensions.size());
	 std::cout<<"NDarray vector TOTAL SIZE = "<< values.size() <<std::endl;
     std::cout<<"NDarray vector[0]...[end] = { "<< values[0] << " , ... , " <<  values[values.size()-1] << " }" <<std::endl;	
     std::cout<<"================================================== " <<std::endl;	 
     return;
   };
// print vector
   inline void PrintVector(std::vector<size_t> vec_, std::size_t size_){		  	  
	  std::cout << " { ";
          for (std::size_t j = 0; j < size_; j++) {
             std::cout << vec_[j];
			 if(j<size_-1) {
				 std::cout <<  " , ";
			 }
          }	  
	  std::cout << " } " << std::endl;	  		
      return;
   };	
   inline void PrintVector(std::vector<double> vec_, std::size_t size_){		  	  
	  std::cout << " { ";
          for (std::size_t j = 0; j < size_; j++) {
             std::cout << vec_[j];
			 if(j<size_-1) {
				 std::cout <<  " , ";
			 }
          }	  
	  std::cout << " } " << std::endl;	  		
      return;
   };     
// Load values of the NDarray from file
    inline void LoadFromFile(std::string filename){
        std::ifstream f;
		
//		std::cout << "filename = " << filename << std::endl;
		
        f.open(filename, std::ifstream::in);
        if (!f.is_open()) {
          std::cout << "Can not open a *.dat file!\n";    
		}
        T r;
// alternative to push_back
//		int i=0; // uncomment when use the alternative
 
		values.clear(); // comment when use the alternative
		
        while (!f.eof())
         {
          f>>r;
// alternative to push_back		  
//		  values[i]=r; // uncomment when use the alternative
//		  i++; // uncomment when use the alternative
          values.push_back(r); // comment when use the alternative
         }
       f.close();
//	   std::cout << "total SIZE =" << values.size() << std::endl;
//	   std::cout << values[0] << "--" << values[1] << "--" << values[values.size()-1] << std::endl;
       return;
	   
   };
/////////////////////////////////////////////////	   
// compute total number of values in the NDarray
    std::size_t computeTotalSize(const std::vector<std::size_t>& dimensions_) const{
    	std::size_t totalSize = 1;

        for (std::size_t i=0; i < dimensions_.size(); i++) {
          totalSize *= dimensions_[i];
		}
//	    for (auto i : dimensions_) {
//          totalSize *= i;
//        }
			
        return totalSize;
    };    
	
  private:

    std::vector<std::size_t> dimensions; // dimensions of the NDarray, example {3,4,3,7} or {100,200,40,50,80}
    std::vector<T> values;  // numbers of the array in the single vector values


}; // end of class NDarray

}// end of namespace ExDiff

#endif /* NDARRAY_HH */
