/*
 * Particle.h
 *
 *  Created on: 17 NOV 2016
 *      Author: Roman Ryutin
 */

#ifndef PARTICLE_HH
#define PARTICLE_HH

#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <cassert>

#include "I.h"
#include "PI.h"

namespace ExDiff{

// aux class for particles definition
class Particle{
  public:
    Particle():
	 p({0.0,0.0,0.0,0.0}),
	 E(0.0),
	 px(0.0),
	 py(0.0),
	 pz(0.0),
	 m(0.0),
	 pp(0.0),
	 pT(0.0),
	 p3D(0.0),
	 ID(0),
	 DSpin(0),
	 Colour(0),
	 Anticolour(0)
	{}
	
    Particle(const std::vector<double>& p_, const double& m_, const int& ID_, const int& DSpin_, const int& Colour_, const int& Anticolour_):
	 p(p_),
	 E(p_[0]),
	 px(p_[1]),
	 py(p_[2]),
	 pz(p_[3]),
	 m(m_),
	 pp(p_[0]*p_[0]-p_[1]*p_[1]-p_[2]*p_[2]-p_[3]*p_[3]),
	 pT(sqrt(p_[1]*p_[1]+p_[2]*p_[2])),
	 p3D(sqrt(p_[1]*p_[1]+p_[2]*p_[2]+p_[3]*p_[3])),
	 ID(ID_),
	 DSpin(DSpin_),
	 Colour(Colour_),
	 Anticolour(Anticolour_)
	{}	
	
//	Particle(Particle& particle_){}
    virtual ~Particle(){}
	
	void Print();
	
	inline void Reset(const std::vector<double>& p_, const double& m_, const int& ID_, const int& DSpin_, const int& Colour_, const int& Anticolour_){
		p = p_;
		E = p_[0];
		px = p_[1];
		py = p_[2];
		pz = p_[3];
		m = m_;
		ID = ID_;
		DSpin = DSpin_;
		Colour = Colour_;
		Anticolour = Anticolour_;
		pp = p_[0]*p_[0]-p_[1]*p_[1]-p_[2]*p_[2]-p_[3]*p_[3];
		pT = sqrt(p_[1]*p_[1]+p_[2]*p_[2]);
		p3D = sqrt(p_[1]*p_[1]+p_[2]*p_[2]+p_[3]*p_[3]);
		return;
	}

	inline void ResetP(const std::vector<double>& p_){
		p = p_;
		E = p_[0];
		px = p_[1];
		py = p_[2];
		pz = p_[3];
		pp = p_[0]*p_[0]-p_[1]*p_[1]-p_[2]*p_[2]-p_[3]*p_[3];
		pT = sqrt(p_[1]*p_[1]+p_[2]*p_[2]);
		p3D = sqrt(p_[1]*p_[1]+p_[2]*p_[2]+p_[3]*p_[3]);
		return;
	}	
	
	inline const std::vector<double>& _p() const{
		return p;
	};
	inline std::vector<double>& _p(){
		return p;
	};	

	inline const double& _px() const{
		return px;
	};
	inline double& _px(){
		return px;
	};	
	
	inline const double& _py() const{
		return py;
	};
	inline double& _py(){
		return py;
	};	
	
	inline const double& _pz() const{
		return pz;
	};
	inline double& _pz(){
		return pz;
	};	

	inline const double& _E() const{
		return E;
	};
	inline double& _E(){
		return E;
	};	
	
	inline const double& _m() const{
		return m;
	};
	inline double& _m(){
		return m;
	};	
	
	inline const double& _pp() const{
		return pp;
	};
	inline double& _pp(){
		return pp;
	};	

	inline const double& _pT() const{
		return pT;
	};
	inline double& _pT(){
		return pT;
	};
	
	inline const double& _p3D() const{
		return p3D;
	};
	inline double& _p3D(){
		return p3D;
	};	

	inline const int& _DSpin() const{
		return DSpin;
	};
	inline int& _DSpin(){
		return DSpin;
	};
	
	inline const int& _ID() const{
		return ID;
	};
	inline int& _ID(){
		return ID;
	};	

	inline const int& _Colour() const{
		return Colour;
	};
	inline int& _Colour(){
		return Colour;
	};
	
	inline const int& _Anticolour() const{
		return Anticolour;
	};
	inline int& _Anticolour(){
		return Anticolour;
	};	
	
	double Rapidity(std::vector<double>&);
	double Pseudorapidity(std::vector<double>&);
    double Theta(std::vector<double>&);	
    double Phi(std::vector<double>&);
	
	int OnShell();

// operations with LVectors that we need to use
//    std::vector<double> Lplus(std::vector<double>&, std::vector<double>&);
//    std::vector<double> Lminus(std::vector<double>&, std::vector<double>&);	
	double Lmult(std::vector<double>&, std::vector<double>&);
	std::vector<double> Lplus(std::vector<double>&, std::vector<double>&);	
	std::vector<double> Lminus(std::vector<double>&, std::vector<double>&);
    void Lprint(std::vector<double>&);	
		
  private:
// m - onshell mass
// p = {E, px, py, pz} - 4-momentum
// pp = E*E - px*px - py*py - pz*pz
   std::vector<double> p;
   double m, pp, pT, p3D;   
   double E, px, py, pz;
// DSpin = Spin * 2 + 1
// ID from PYTHIA 
// Colour = 0 (colourless), 1,2,3 (R G B)  
// Anticolour = 0 (colourless), 1,2,3 (antiR antiG antiB) 
   int ID, DSpin, Colour, Anticolour;
  
//		
	
}; // end of class Particle definitions

}// end of namespace ExDiff

#endif /* PARTICLE_HH */