// running generator alone without external packages
// compilation:
// make clean
// make ExDiff
// ExDiff config/cardname.card > outfile

#include <fstream>
#include <cassert>
#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <stdio.h>
#include <time.h>
#include <iomanip>
#include <limits>

#include <string>
#include <ctime>

#include "inc/I.h"
#include "inc/PI.h"
#include "inc/newrand.h"
#include "inc/NDarray.h"
#include "inc/NDgrid.h"
#include "inc/iFunc.h"
#include "inc/Generator.h"
#include "inc/Particle.h"
#include "inc/Constants.h"
#include "inc/Kinematics.h"
#include "inc/Event.h"
#include "inc/Interface.h"

using namespace ExDiff;

int main(int argc, const char** argv){
	   
// final construction
// datacard is the argument of the main file (filename)

   const char* datacard = argv[1]; 

// reading input data card
// create interface   
   Interface iogen(datacard);
   iogen.GetCard(datacard);
   iogen.GeneratorInfo();
   iogen.PrintPars();
      
// generator initialization   
   std::cout << "Initialization of ExDiff table ..................." << std::endl;
   Generator gen(iogen.GeneratorInit().GetFUN(),iogen.GeneratorInit().GetFUNG(),iogen.GeneratorInit().GetFUNGTOT());
   KinCM kin(iogen.KinematicsInit(gen)._sqs(),iogen.KinematicsInit(gen)._particles());
   Event ev(iogen.EventInit(kin)._particles());

//---------------

// for output=1,2,3 (PYTHIA8, ROOT, HEPMC etc.)
  if (iogen._outformat() > 0 && iogen._outformat() < 10){

     std::cout << "Using this card file, you have to link external packages," << std::endl;
     std::cout << "such as PYTHIA, ROOT, HEPMC etc. " << std::endl;
	 std::cout << "or change the output format in the card file" << std::endl;
     std::cout << "Now generator is writing the usual file *.exdiff for the process" << std::endl;
     iogen.GenerateFile(gen,kin,ev);
  }

  if (iogen._outformat() == 0 ) {
// for intrinsic output format

   std::cout << "Writing to the *.exdiff file for the process ............" << std::endl;
   iogen.GenerateFile(gen,kin,ev);
   
  }
  
// Analizers can be added (see mainExDiff.cc)   
  if (iogen._outformat() >= 10 ) {
// for intrinsic output format
   
   std::cout << "CUTS TO CHECK ++++++++++++++++++++++++++++++++++" << std::endl;
   std::cout << "SETVAR = " << iogen.SETVAR << "   SETCUT = " << iogen.SETCUT << std::endl;
   std::cout << "Writing to *.exdiff* files for the process with cuts .........." << std::endl;
   iogen.GenerateFileCuts(gen,kin,ev);
   
  }

//---------------

return 0;

}
