// Example for working with PYTHIA8.2 output and hadronization
// Some pythia code could be removed, if you do not need it
// This one generate pythia events like elastic and CEP with hadrons (without QCD partons)
// compilation:
// make clean
// make ExDiffPY
// ExDiffPY config/cardname.card > outfile

#include <fstream>
#include <cassert>
#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <stdio.h>
#include <time.h>
#include <iomanip>
#include <limits>

#include <string>
#include <ctime>

#include "inc/I.h"
#include "inc/PI.h"
#include "inc/newrand.h"
#include "inc/NDarray.h"
#include "inc/NDgrid.h"
#include "inc/iFunc.h"
#include "inc/Generator.h"
#include "inc/Particle.h"
#include "inc/Constants.h"
#include "inc/Kinematics.h"
#include "inc/Event.h"
#include "inc/Interface.h"

#include "Pythia8/Pythia.h"

using namespace Pythia8;

//==========================================================================
//====== additional functions to set up events =============================
// filling the pythia8 event with hadrons only
void fillEventHadrons(ExDiff::Event& ev, Event& event, ParticleData& pdt, Rndm& rndm) {

  // Reset event record to allow for new event.
  event.reset();

// ev._particles().size() = number of particles in the event
// ev._particles()[i]._ID() = ID of the	particle
// ev._particles()[i]._m() = mass of the particle
// ev._particles()[i]._E() = energy of the particle
// ev._particles()[i]._px() = px of the	particle
// ev._particles()[i]._py() = py of the	particle
// ev._particles()[i]._pz() = pz of the	particle	

// === elastic scattering of hadrons ===
  if(ev._particles().size() == 4){
// Add beam protons to the event
    event.append( ev._particles()[0]._ID(), -12, 0, 0, 3, 0, ev._particles()[0]._Colour(), ev._particles()[0]._Anticolour(), ev._particles()[0]._px(), ev._particles()[0]._py(), ev._particles()[0]._pz(), ev._particles()[0]._E(), ev._particles()[0]._m());
    event.append( ev._particles()[1]._ID(), -12, 0, 0, 4, 0, ev._particles()[1]._Colour(), ev._particles()[1]._Anticolour(), ev._particles()[1]._px(), ev._particles()[1]._py(), ev._particles()[1]._pz(), ev._particles()[1]._E(), ev._particles()[1]._m());
// add scattered hadron 1
    event.append( ev._particles()[2]._ID(), 14, 1, 0, 0, 0, ev._particles()[2]._Colour(), ev._particles()[2]._Anticolour(), ev._particles()[2]._px(), ev._particles()[2]._py(), ev._particles()[2]._pz(), ev._particles()[2]._E(), ev._particles()[2]._m());
// add scattered proton 2
    event.append( ev._particles()[3]._ID(), 14, 2, 0, 0, 0, ev._particles()[3]._Colour(), ev._particles()[3]._Anticolour(), ev._particles()[3]._px(), ev._particles()[3]._py(), ev._particles()[3]._pz(), ev._particles()[3]._E(), ev._particles()[3]._m());  
  }

// === CEP with 1 central unstable particle ===
  if(ev._particles().size() == 5){
// Add beam protons to the event
    event.append( ev._particles()[0]._ID(), -12, 0, 0, 3, 4, ev._particles()[0]._Colour(), ev._particles()[0]._Anticolour(), ev._particles()[0]._px(), ev._particles()[0]._py(), ev._particles()[0]._pz(), ev._particles()[0]._E(), ev._particles()[0]._m());
    event.append( ev._particles()[1]._ID(), -12, 0, 0, 4, 5, ev._particles()[1]._Colour(), ev._particles()[1]._Anticolour(), ev._particles()[1]._px(), ev._particles()[1]._py(), ev._particles()[1]._pz(), ev._particles()[1]._E(), ev._particles()[1]._m());
// add scattered proton 1
    event.append( ev._particles()[2]._ID(), 15, 1, 0, 0, 0, ev._particles()[2]._Colour(), ev._particles()[2]._Anticolour(), ev._particles()[2]._px(), ev._particles()[2]._py(), ev._particles()[2]._pz(), ev._particles()[2]._E(), ev._particles()[2]._m());   
// Add the central particle to the event record.
  int iNew = event.append( ev._particles()[3]._ID(), 1, 1, 2, 0, 0, ev._particles()[3]._Colour(), ev._particles()[3]._Anticolour(), ev._particles()[3]._px(), ev._particles()[3]._py(), ev._particles()[3]._pz(), ev._particles()[3]._E(), ev._particles()[3]._m());
// Generate lifetime, to give decay away from primary vertex.
  event[iNew].tau( event[iNew].tau0() * rndm.exp() );
// add scattered proton 2
    event.append( ev._particles()[4]._ID(), 15, 2, 0, 0, 0, ev._particles()[4]._Colour(), ev._particles()[4]._Anticolour(), ev._particles()[4]._px(), ev._particles()[4]._py(), ev._particles()[4]._pz(), ev._particles()[4]._E(), ev._particles()[4]._m());   
  }

// === CEP with 2 central unstable particles ===
  if(ev._particles().size() == 6){
// Add beam protons to the event
    event.append( ev._particles()[0]._ID(), -12, 0, 0, 3, 4, ev._particles()[0]._Colour(), ev._particles()[0]._Anticolour(), ev._particles()[0]._px(), ev._particles()[0]._py(), ev._particles()[0]._pz(), ev._particles()[0]._E(), ev._particles()[0]._m());
    event.append( ev._particles()[1]._ID(), -12, 0, 0, 5, 6, ev._particles()[1]._Colour(), ev._particles()[1]._Anticolour(), ev._particles()[1]._px(), ev._particles()[1]._py(), ev._particles()[1]._pz(), ev._particles()[1]._E(), ev._particles()[1]._m());

// add scattered proton 1
    event.append( ev._particles()[2]._ID(), 15, 1, 0, 0, 0, ev._particles()[2]._Colour(), ev._particles()[2]._Anticolour(), ev._particles()[2]._px(), ev._particles()[2]._py(), ev._particles()[2]._pz(), ev._particles()[2]._E(), ev._particles()[2]._m());    
// Add the first central particle to the event record.
  int iNew1 = event.append( ev._particles()[3]._ID(), 1, 1, 0, 0, 0, ev._particles()[3]._Colour(), ev._particles()[3]._Anticolour(), ev._particles()[3]._px(), ev._particles()[3]._py(), ev._particles()[3]._pz(), ev._particles()[3]._E(), ev._particles()[3]._m());
// Generate lifetime, to give decay away from primary vertex.
  event[iNew1].tau( event[iNew1].tau0() * rndm.exp() );
// Add the first central particle to the event record.
  int iNew2 = event.append( ev._particles()[4]._ID(), 1, 2, 0, 0, 0, ev._particles()[4]._Colour(), ev._particles()[4]._Anticolour(), ev._particles()[4]._px(), ev._particles()[4]._py(), ev._particles()[4]._pz(), ev._particles()[4]._E(), ev._particles()[4]._m());
// Generate lifetime, to give decay away from primary vertex.
  event[iNew2].tau( event[iNew2].tau0() * rndm.exp() );
// add scattered proton 2
    event.append( ev._particles()[5]._ID(), 15, 2, 0, 0, 0, ev._particles()[5]._Colour(), ev._particles()[5]._Anticolour(), ev._particles()[5]._px(), ev._particles()[5]._py(), ev._particles()[5]._pz(), ev._particles()[5]._E(), ev._particles()[5]._m());   
  }

}

// MAIN function
int main(int argc, const char** argv){

// BASIC PART ------ begin
// final construction
// datacard is the argument of the main file (filename)
   const char* datacard = argv[1]; 

// reading input data card
// create interface   
   ExDiff::Interface iogen(datacard);
   iogen.GetCard(datacard);
   iogen.GeneratorInfo();
   iogen.PrintPars();
      
// generator initialization   
   std::cout << "Initialization of ExDiff table ..................." << std::endl;
   ExDiff::Generator gen(iogen.GeneratorInit().GetFUN(),iogen.GeneratorInit().GetFUNG(),iogen.GeneratorInit().GetFUNGTOT());
   ExDiff::KinCM kin(iogen.KinematicsInit(gen)._sqs(),iogen.KinematicsInit(gen)._particles());
   ExDiff::Event ev(iogen.EventInit(kin)._particles());
// BASIC PART ------ end 

// for output=1,2,3 (PYTHIA8, ROOT, HEPMC etc.)
  if (iogen._outformat() > 0 && iogen._outformat() < 10){

  // Set number of events to generate and to list.
  int nEvent = iogen._Nevents();
  int nList = 3;
  int type = 0;

  // Generator; shorthand for event and particleData.
  Pythia pythia;

  // Key requirement: switch off ProcessLevel, and thereby also PartonLevel.
  pythia.readString("ProcessLevel:all = off");

  // Optionally switch off resonance decays, or only showers in them.
  //pythia.readString("ProcessLevel:resonanceDecays = off");
  //pythia.readString("PartonLevel:FSRinResonances = off");

  // Optionally switch off ordinary decays.
  //pythia.readString("HadronLevel:Decay = off");

  // Switch off automatic event listing in favour of manual.
  pythia.readString("Next:numberShowInfo = 0");
  pythia.readString("Next:numberShowProcess = 0");
  pythia.readString("Next:numberShowEvent = 0");

  // Set true to also see space-time information in event listings.
  bool showScaleAndVertex = (type == 0) ? true : false;

  // Initialize.
  pythia.init();

// Generator; shorthand for event and particleData.
  Event& event      = pythia.event;
  ParticleData& pdt = pythia.particleData; 
  
  // Book histograms. Example.
  double ee = 20.0;
  Hist epCons("deviation from energy-momentum conservation",100,0.,1e-4);
  Hist chgCons("deviation from charge conservation",57,-9.5,9.5);
  Hist nFinal("final particle multiplicity",100,-0.5,99.5);
  Hist dnparticledp("dn/dp for particles",100,0.,ee);
  Hist status85("multiplicity status code 85",50,-0.5,49.5);
  Hist status86("multiplicity status code 86",50,-0.5,49.5);
  Hist status83("multiplicity status code 83",50,-0.5,49.5);
  Hist status84("multiplicity status code 84",50,-0.5,49.5);
  Hist dndtheta("particle flow in event plane",100,-M_PI,M_PI);
  Hist dedtheta("energy flow in event plane",100,-M_PI,M_PI);
  Hist dpartondtheta("parton flow in event plane",100,-M_PI,M_PI);
  Hist dndyAnti("dn/dy primaries antijunction",100, -10., 10.);
  Hist dndyJun("dn/dy primaries junction",100, -10., 10.);
  Hist dndySum("dn/dy all primaries",100, -10., 10.);

  // Begin of event loop.
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    ev = iogen.GenerateEvent(gen,kin,ev);
    fillEventHadrons( ev, event, pdt, pythia.rndm);

    // Generate events. Quit if failure.
    if (!pythia.next()) {
      cout << " Event generation aborted prematurely, owing to error!\n";
      break;
    }

    // List first few events.
    if (iEvent < nList) {
      event.list(showScaleAndVertex);
      // Also list junctions.
      event.listJunctions();
    }

	// Example ------------------------------------------------
    // Initialize statistics.
    Vec4 pSum = - event[0].p();
    double chargeSum = 0.;
    chargeSum = -event[1].charge();
    
    int nFin = 0;
    int n85 = 0;
    int n86 = 0;
    int n83 = 0;
    int n84 = 0;

    // Loop over all particles.
    for (int i = 0; i < event.size(); ++i) {
      int status = event[i].statusAbs();

      // Find any unrecognized particle codes.
      int id = event[i].id();
      if (id == 0 || !pdt.isParticle(id))
        cout << " Error! Unknown code id = " << id << "\n";

      // Find particles with E-p mismatch.
      double eCalc = event[i].eCalc();
      if (abs(eCalc/event[i].e() - 1.) > 1e-6) cout << " e mismatch, i = "
        << i << " e_nominal = " << event[i].e() << " e-from-p = "
        << eCalc << " m-from-e " << event[i].mCalc() << "\n";

      // Parton flow in event plane.
      if (status == 71 || status == 72) {
        double thetaXZ = event[i].thetaXZ();
        dpartondtheta.fill(thetaXZ);
      }

      // Origin of primary hadrons.
      if (status == 85) ++n85;
      if (status == 86) ++n86;
      if (status == 83) ++n83;
      if (status == 84) ++n84;

      // Flow of primary hadrons in the event plane.
      if (status > 80 && status < 90) {
        double eAbs = event[i].e();
        if (eAbs < 0.) {cout << " e < 0 line " << i; event.list();}
        double thetaXZ = event[i].thetaXZ();
        dndtheta.fill(thetaXZ);
        dedtheta.fill(thetaXZ, eAbs);

        // Rapidity distribution of primary hadrons.
        double y = event[i].y();
        dndySum.fill(y);
      }

      // Study final-state particles.
      if (event[i].isFinal()) {
        pSum += event[i].p();
        chargeSum += event[i].charge();
        nFin++;
        double pAbs = event[i].pAbs();
        dnparticledp.fill(pAbs);
      }
    }

    // Fill histograms once for each event.
    double epDev = abs(pSum.e()) + abs(pSum.px()) + abs(pSum.py())
      + abs(pSum.pz());
    epCons.fill(epDev);
    chgCons.fill(chargeSum);
    nFinal.fill(nFin);
    status85.fill(n85);
    status86.fill(n86);
    status83.fill(n83);
    status84.fill(n84);
    if (epDev > 1e-3  || abs(chargeSum) > 0.1) event.list();
	
  // End of event loop.
  }

  // Print statistics, histograms and done.
  pythia.stat();
  cout << epCons << chgCons << nFinal << dnparticledp
       << dndtheta << dedtheta << dpartondtheta << dndySum;

  // Done.

  }

// Standardt part for outputs without external packages
  if (iogen._outformat() == 0 ) {
// for intrinsic output format
   std::cout << "Standardt *.exdiff file output ......." << std::endl;
   iogen.GenerateFile(gen,kin,ev);
  }
  
// Analizers can be added (see mainExDiff.cc)   
  if (iogen._outformat() >= 10 ) {
// for intrinsic output format
   std::cout << "Standardt *.exdiff* files output with cuts ......." << std::endl;   
   std::cout << "CUTS TO CHECK ++++++++++++++++++++++++++++++++++" << std::endl;
   std::cout << "SETVAR = " << iogen.SETVAR << "   SETCUT = " << iogen.SETCUT << std::endl;
   iogen.GenerateFileCuts(gen,kin,ev);
  }
  
  return 0;
}
