/*
 * Interface.cpp
 *
 *  Created on: 17 NOV 2016
 *      Author: Roman Ryutin
 */
#include <fstream>
#include <cassert>
#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <stdio.h>
#include <time.h>
#include <iomanip>
#include <limits>
#include <string>
#include <ctime>

#include "../inc/I.h"
#include "../inc/PI.h"
#include "../inc/newrand.h"
#include "../inc/NDarray.h"
#include "../inc/NDgrid.h"
#include "../inc/iFunc.h"
#include "../inc/Generator.h"
#include "../inc/Particle.h"
#include "../inc/Constants.h"
#include "../inc/Kinematics.h"
#include "../inc/Event.h"
#include "../inc/Interface.h"

// functions of class Interface ===========================================
///////////////////////////////////////////////////////////////////////////
void ExDiff::Interface::PrintPars(){ 
	std::cout << "========= Parameters for the generation ===========" <<std::endl;	
	std::cout << "Authors: " << Get_Authors(IDauthors) <<std::endl;
	std::cout << "Process: " << Get_Process(IDprocess) <<std::endl;	
	std::cout << "C.M. energy: " << Get_Energy(IDenergy) << " GeV" <<std::endl;	
	std::cout << "version: " << version << " format: " << kinformat <<std::endl;
	std::cout << "Number of events: " << Nevents << " Output format: " << outformat <<std::endl; 
	std::cout << "===================================================" <<std::endl;	
//    printf("TOTAL C.-S. = %.6f mb \n", CSTOT);	
	return; 
};


// Basic settings
void ExDiff::Interface::GetCard(const char *_filename){
	std::ifstream f; 
	 
	std::vector<int> pars;
	pars.clear();

    f.open(_filename, std::ifstream::in);
    if (!f.is_open()) {
       std::cout << "Can not open a *.card file!\n";   
	}	
	
	int par = 0;
	for(int i = 0; i < 7; i++){
	  f >> par;
      pars.push_back(par);	  
	}
// IDauthors IDprocess IDenergy version Nevents kinformat outformat	
    IDauthors = pars[0];
	IDprocess = pars[1];
	IDenergy = pars[2];
	version = pars[3];
	Nevents = pars[4];
	kinformat = pars[5];
	outformat = pars[6];	
	
// set up SETVAR, SETCUT
    if(outformat >= 10 && outformat < 20){SETVAR = 1;}
    if(outformat >= 20 && outformat < 30){SETVAR = 2;}
    if(outformat >= 30 && outformat < 40){SETVAR = 3;}
    if(outformat >= 40 && outformat < 50){SETVAR = 4;}	
	if(outformat == 11 || outformat == 21 || outformat == 41){SETCUT = 1;}
	if(outformat == 31){SETCUT = 4;}
	if(outformat == 12 || outformat == 22 || outformat == 32 || outformat == 42){SETCUT = 2;}
	if(outformat == 13 || outformat == 23 || outformat == 33 || outformat == 43){SETCUT = 3;}
    
	f.close();
	
	pars.clear();

	std::cout << "GetCard ---------------------------"  << std::endl; 
	
	return;
};	

// Init masses for the use in Exdiff class
std::vector<double> ExDiff::Interface::MassesInit(){	

	std::cout << "MassesInit ---------------------------"  << std::endl; 

    ExDiff::Constants c;
    double mc1_ = c.m_pi;
    double mc2_ = c.m_pi;
    double mc_ = 1.0; 
	double m = c.m_p;
	
	std::vector<double> Masses;
	
	Masses.clear();
	Masses.push_back(m);

      if(IDprocess == 1){ 
	    mc_ = 1.72; 
	  }  
	  
      if(IDprocess == 2 || IDprocess == 20 || IDprocess == 21 || IDprocess == 22){ 
       mc_ = 1.95; 
	  }
	  
      if(IDprocess == 5 || IDprocess == 50 || IDprocess == 51 || IDprocess == 52){ 
       mc_ = 1.2751;   
	  }	  
	  
      if(IDprocess == 6 || IDprocess == 60 || IDprocess == 61 || IDprocess == 62){ 
       mc_ = 2.22; 
	  }		  
	  
      if(IDprocess == 4){ 
	    mc_ = 1.5; 
	  }
      	  
	  if(IDprocess == 3){  
		Masses.push_back(mc1_);
		Masses.push_back(mc2_);
		Masses.push_back(m);	
	  }	
       
      if(IDprocess == 1 || IDprocess == 2 || IDprocess == 4 || IDprocess == 5 || IDprocess == 6 || IDprocess > 19){
		Masses.push_back(mc_);
		Masses.push_back(m);	  	
	  }
	  
      if(IDprocess == 0){
		Masses.push_back(m);		  
	  }	
	     	
    return Masses;	
};


// Init IDs for the use in Exdiff class
std::vector<int> ExDiff::Interface::IDsInit(){	
	std::cout << "IDsInit ---------------------------"  << std::endl; 
    ExDiff::Constants c;
    int id_ = 0; 
	int id1_ = 0; 
	int id2_ = 0; 
    int idp = c.IDproton;	
	
	std::vector<int> IDs({0});

	IDs.clear();
	IDs.push_back(idp);

      if(IDprocess == 1){  
		id_ = c.ID1710;
	  }  
	  
      if(IDprocess == 2 || IDprocess == 20 || IDprocess == 21 || IDprocess == 22){ 
	   id_ = c.ID1950;   
	  }
	  
      if(IDprocess == 5 || IDprocess == 50 || IDprocess == 51 || IDprocess == 52){ 
	   id_ = c.ID1270;   
	  }	  
	  
      if(IDprocess == 6 || IDprocess == 60 || IDprocess == 61 || IDprocess == 62){ 
	   id_ = c.ID2220; 
	  }		  
	  
      if(IDprocess == 4){ 
		id_ = c.ID1500;
	  }

	  if(IDprocess == 3){  
		id1_ = c.IDpiplus;
		id2_ = c.IDpiminus; 
		IDs.push_back(id1_);
		IDs.push_back(id2_);
		IDs.push_back(idp);		
	  }	
       
      if(IDprocess == 1 || IDprocess == 2 || IDprocess == 4 || IDprocess == 5 || IDprocess == 6 || IDprocess > 19){
		IDs.push_back(id_);
		IDs.push_back(idp);		  	
	  }
	  
      if(IDprocess == 0){
		IDs.push_back(idp);			  
	  }	
	     	
    return IDs;	
};

// Generator initialization function -------------------------------------
	
ExDiff::Generator ExDiff::Interface::GeneratorInit(){
	std::cout << "GeneratorInit ---------------------------"  << std::endl; 	
	int flag = 0; // grid from file, if !=0 - special version of the grid
    int i = 0;
	ExDiff::Constants c;
	
	std::vector<std::size_t> dim;
	double aux = 0.0;
	
// create NDarray, NDgrid using parameters IDauthors, IDprocess, version 
   if(IDauthors == 1){

      if(IDprocess == 1  || IDprocess == 2 || IDprocess == 4 || IDprocess == 5 || IDprocess == 6 || IDprocess > 19){
        if(version == 1 || version == 2){
           dim = {129,129,9,9};
		   flag = 1; // uniform grid version 1 by cycle
        }  		
// example of other versions, 1 is default version	  
//      if(version == 2){
//         std::vector<std::size_t> dim({400,400,20,20});
//      }    
	   
	  }   
	  if(IDprocess == 0){
         if(version == 1){
           dim = {501};
         } 			
	  }
	   	   
   }

   if(IDauthors == 2){

      if(IDprocess == 3 ){
        if(version == 1){
           dim = {15,15,6,15,15,8,6};
		   flag = 1; // nonuniform grid version 4
        }  		   
	   
	  }   
   	   
   }
   
   ExDiff::NDarray<double> dat_(dim);
   ExDiff::NDgrid<double>  grid_(dim);
	   
// create names of dat and grid files
   std::string datfile = DatFileName();    

// read NDarray, NDgrid from files, or, NDgrid create inside this function
//   std::cout << "dat file = " << datfile << std::endl;
   dat_.LoadFromFile(datfile);
   
   double mc1_ = c.m_pi;
   double mc2_ = c.m_pi;
   double mc_ = 1.0; 
   int id_ = 0;
   int id1_ = 0; 
   int id2_ = 0;   
   double sqs_ = Get_Energy(IDenergy); 
   double Mcmax_;

   std::string gridfile = GridFileName();  
//   std::cout << "grid file = " << gridfile << std::endl;   
   
   if(flag != 0){
//	  printf("Fill Grid, flag = %1i - 1", flag); 
// fill the uniform grid with numbers
     if(flag == 1){
		 
      double ximin_ = 0.0001;

      if(IDprocess == 1){ 
	    mc_ = 1.72; 
		id_ = c.ID1710;
        grid_ = grid_2(mc_, ximin_, sqs_);   
	  }  
	  
      if(IDprocess == 2 || IDprocess == 20 || IDprocess == 21 || IDprocess == 22){ 
       mc_ = 1.95; 
	   id_ = c.ID1950;
//	   printf("Fill Grid, flag = %1i - 2\n", flag); 
       grid_ = grid_2(mc_, ximin_, sqs_ ); 
//   	   printf("Fill Grid, flag = %1i - 3\n", flag); 	   
	  }
	  
      if(IDprocess == 5 || IDprocess == 50 || IDprocess == 51 || IDprocess == 52){ 
       mc_ = 1.2751; 
	   id_ = c.ID1270;
       grid_ = grid_2(mc_, ximin_, sqs_ ); 	   
	  }	  
	  
      if(IDprocess == 6 || IDprocess == 60 || IDprocess == 61 || IDprocess == 62){ 
       mc_ = 2.22; 
	   id_ = c.ID2220;
       grid_ = grid_2(mc_, ximin_, sqs_ ); 	   
	  }		  
	  
      if(IDprocess == 4){ 
	    mc_ = 1.5; 
		id_ = c.ID1500;
        grid_ = grid_2(mc_, ximin_, sqs_);   
	  }

	  if(IDprocess == 3){  
		id1_ = c.IDpiplus;
		id2_ = c.IDpiminus;
		Mcmax_ = 3.0;
        grid_ = grid_4(mc1_, mc2_);   
	  }

	  
     }	  
	   
   } else{    
//      printf("Load Grid From File, flag = %1i", flag);
      grid_.LoadFromFile(gridfile);
   } 

//    printf("Fill Grid, flag = %1i - 4\n", flag); 
//	grid_.printC(); 
// create iFunc
    ExDiff::iFunc fungen(dat_,grid_);
// create Generator	
    ExDiff::Generator gen_(fungen); 
    ExDiff::newrand();

	  if(IDprocess != 3){
//	    printf("TOTAL C-S. = %.6f mb \n", gen_.GetFUNGTOT());
		CSTOT = gen_.GetFUNGTOT();
	  }
	  else{
//	    printf("TOTAL C-S. = %.6f mub \n", 2.0*1000.0*gen_.GetFUNGTOT());
        CSTOT = 2.0*gen_.GetFUNGTOT();	
	  }

    return gen_;	
};	
	

// Kinematics initialization function ====================


ExDiff::KinCM ExDiff::Interface::KinematicsInit(ExDiff::Generator& gen_){
	std::cout << "KinematicsInit ---------------------------"  << std::endl; 	
// generate 1 sample
    std::vector<double> genvec;	
	genvec = gen_.Generate();
// additional phi variable;	
//    fungen.PrintVector(genvec,genvec.size()); 

	genvec.push_back(2.0*PI*gen_.RNUM());
	
// create KinCM object
// here we use default set of variables
// 2 -> 2 {t}; phi
// 2 -> 3 {DeltaT_1, DeltaT_2, phi_12, log(xi_1)}; phi_1
// 2 -> 4 {D1, D2, phi_12, Mc, V_yc; V_Eta_j, Phi_j}; phi_1;
    std::vector<double> masses__ = MassesInit();
	std::vector<int> ids__ = IDsInit();  	
    
	double aux = 0.0;
    if(IDprocess == 3){  	
//	  masses__ = {c.m_p,mc1_,mc2_,c.m_p};
//	  ids__ = {c.IDproton,id1_,id2_,c.IDproton};	  	
// Additional symmetry random changes in Eta_j
// for process 2 to 4, kinematics ID=0
// If RND<0.5 genvec[5] -> -genvec[5];	
      if(gen_.RNUM() > 0.5) {
		  aux = genvec[5];
		  genvec[5] = - aux;
	  }		  

	}	
		
	double sqs_ = Get_Energy(IDenergy); 
	
	ExDiff::KinCM kinematics(sqs_,0,genvec,masses__,ids__);
		
	return kinematics;
}; 

// Event initialization function
ExDiff::Event ExDiff::Interface::EventInit(ExDiff::KinCM& kin_){
	std::cout << "EventInit ---------------------------"  << std::endl; 
    ExDiff::Event ev(kin_.EvParticles());
    return ev;	
};  

// Event generation
ExDiff::Event ExDiff::Interface::GenerateEvent(ExDiff::Generator& gen_, ExDiff::KinCM& kin_, ExDiff::Event& ev_){

    double aux = 0.0;
    std::vector<double> genvec;	
	
	genvec = gen_.Generate(); 	
	genvec.push_back(2.0*PI*gen_.RNUM());
	
    if(IDprocess == 3){
// Additional symmetry random changes in Eta_j
// for process 2 to 4, kinematics ID=0
// If RND<0.5 genvec[5] -> -genvec[5];	
      if(gen_.RNUM() > 0.5){
		  aux = genvec[5];
		  genvec[5] = - aux;
	  }	
		}
		
	kin_.ResetVars(genvec); 			 
	ev_.Reset(kin_.EvParticles());	
	
	return ev_;	
};


// generate file of events in case of 0 format,
void ExDiff::Interface::GenerateFile(ExDiff::Generator& gen_, ExDiff::KinCM& kin_, ExDiff::Event& ev_){
	std::cout << "GenerateFile ---------------------------"  << std::endl; 	
	double aux = 0.0;
	std::vector<double> genvec;
	std::string outputfile = ResultFileName();

// time variables ////////////////////////////////////////
    std::clock_t start;
    double duration;
/////////////////////////////////////////////////////////

// for simple output outformat=0 and all events are collected
// to the output file
// for PYTHIA and HEPMC output (outformat>0)
// function returns single event

// Begin BLOCK of output	
  if(outformat == 0){

	for(int i = 0; i < Nevents; i++){

// time for calculations start	//////////////////////////
			if(i == 0){		
              std::cout << "clock start = = " << std::clock() << std::endl;			
			  start = std::clock();
			}
/////////////////////////////////////////////////////////		
		
		    genvec.clear();
		    genvec = gen_.Generate(); 
	        genvec.push_back(2.0*PI*gen_.RNUM());
		
            if(IDprocess == 3){
// Additional symmetry random changes Eta_j
// for process 2 to 4, kinematics ID=0
// If RND<0.5 genvec[5] -> -genvec[5];	
              if(gen_.RNUM() > 0.5) {
		        aux = genvec[5];
		        genvec[5] = - aux;
	          }	

		    }
		
		    kin_.ResetVars(genvec); 	
		    ev_.Reset(kin_.EvParticles());

// FOR DEVELOPERS ================================================================================
// You can use directly all the parameters of particles in the event
// to make your own output or connection to your software
// instead of the string 

// remove after new changes		
		    ev_.AddToFile(outformat,outputfile);				
		
// use your own method with the following parameters of all particles in the event
// i: 0 => ev._particles().size()-1 [i changes 0=>3 (2 to 2), 0=>4 (2 to 3) and 0=>5 (2 to 4)]
// ev._particles().size() = number of particles in the event
// ev._particles()[i]._ID() = ID of the	particle
// ev._particles()[i]._m() = mass of the	particle
// ev._particles()[i]._E() = energy of the	particle
// ev._particles()[i]._px() = px of the	particle
// ev._particles()[i]._py() = py of the	particle
// ev._particles()[i]._pz() = pz of the	particle	
// FOR DEVELOPERS ================================================================================
		
// time for calculations end ///////////////////////////////////////////////			 				
            if(i == Nevents - 1){ 
			std::cout<<"clock stop = "<< std::clock() << '\n';
            duration = ( std::clock() - start ) / (double) CLOCKS_PER_SEC;
            std::cout<<"duration = "<< duration << '\n';				
			}		
////////////////////////////////////////////////////////////////////////////		

		
	}
  }
  else {

		genvec.clear();		
		genvec = gen_.Generate(); 		
	    genvec.push_back(2.0*PI*gen_.RNUM());

		
        if(IDprocess == 3){
// Additional symmetry random changes in Eta_j
// for process 2 to 4, kinematics ID=0
// If RND<0.5 genvec[5] -> -genvec[5];	
      if(gen_.RNUM() > 0.5) {
		  aux = genvec[5];
		  genvec[5] = - aux;
	  }	

		}
		
		kin_.ResetVars(genvec); 	 
		ev_.Reset(kin_.EvParticles());

// FOR DEVELOPERS ================================================================================
// You can use directly all the parameters of particles in the event
// to make your own output or connection to your software
// instead of the string 			
		
// use your own method with the following parameters of all particles in the event
// i: 0 => ev._particles().size()-1 [i changes 0=>3 (2 to 2), 0=>4 (2 to 3) and 0=>5 (2 to 4)]
// ev._particles().size() = number of particles in the event
// ev._particles()[i]._ID() = ID of the	particle
// ev._particles()[i]._m() = mass of the	particle
// ev._particles()[i]._E() = energy of the	particle
// ev._particles()[i]._px() = px of the	particle
// ev._particles()[i]._py() = py of the	particle
// ev._particles()[i]._pz() = pz of the	particle	
// FOR DEVELOPERS ================================================================================

  }

// End BLOCK of output	

	std::cout << "===================================================" <<std::endl;	
    printf("TOTAL C.-S. = %.6f mb \n", CSTOT);

    return;	
	
};
/////////////////////////////////////////////////////////////////////////////////	

// generate file of events in case of 0 format,
// and write to files for the generated events satisfying the cuts
void ExDiff::Interface::GenerateFileCuts(ExDiff::Generator& gen_, ExDiff::KinCM& kin_, ExDiff::Event& ev_){
	
	double aux = 0.0;
	std::vector<double> genvec;
	std::string xfile = ResultFileName();
	std::string oxfile = xfile;
	
	int ivar = SETVAR;
	int icut = SETCUT;		
	int icheck = 1;

    std::cout << "GenerateFileCuts is running ++++++++++++++++++++++++++++++++++" << std::endl;	
    std::cout << "SETVAR = " << SETVAR << "   SETCUT = " << SETCUT << std::endl; 	
	
    if(IDprocess == 0){
	  std::cout << "for elastic scattering variables t1->t, Mc->0, phi0->phi, yc->y !!!"<< std::endl;
	}
	
// filenames for output variables
// t-distribution (generates simple txt file with t-values)
	if (outformat >= 10 && outformat < 20){
		oxfile += "_t";
		ivar = 1;
		}
// Mc-distribution for 2 to 4 CEP process (generates simple txt file with Mc values)	
	if (outformat >= 20 && outformat < 30){
		oxfile += "_Mc";
		ivar = 2;
		}
	if (outformat >= 30 && outformat < 40){
		oxfile += "_phi0";
		ivar = 3;
		}
	if (outformat >= 40 && outformat < 50){
		oxfile += "_yc";
		ivar = 4;
		}	
// output of variables with additional cuts
//  200 GeV STAR pi pi CEP cuts  
    if (outformat == 11 || outformat == 21 || outformat == 41){
		oxfile += "_200";
		xfile += "_200";
		icut = 1;
		}		
    if (outformat == 31){
		oxfile += "_200";
		xfile += "_200";
		icut = 4;
		}		
//  1960 GeV CDF pi pi CEP cuts  
    if (outformat == 12 || outformat == 22 || outformat == 32 || outformat == 42){
		oxfile += "_1960";
		xfile += "_1960";
		icut = 2;
		}
// 8 TeV LHC cuts pi pi	version 1
    if (outformat == 13 || outformat == 23 || outformat == 33 || outformat == 43){
		oxfile += "_8000_1";
		xfile += "_8000_1";
		icut = 3;
		}
	
    std::cout << "file of events = " << xfile << std::endl;	
	std::cout << " file of the variable = " << oxfile << std::endl;
	
	std::ofstream f;
	double vartofile = 0.0;


	
// Begin BLOCK of output	
    int Ncutevents = 0;

	for(int i = 0; i < Nevents; i++){
		
		    genvec.clear();
		    genvec = gen_.Generate(); 
	        genvec.push_back(2.0*PI*gen_.RNUM());
		
            if(IDprocess == 3){
// Additional symmetry random changes Eta_j
// for process 2 to 4, kinematics ID=0
// If RND<0.5 genvec[5] -> -genvec[5];	
              if(gen_.RNUM() > 0.5) {
		        aux = genvec[5];
		        genvec[5] = - aux;
	          }	

		    }
		
		    kin_.ResetVars(genvec); 	
		    ev_.Reset(kin_.EvParticles());
			
			icheck = kin_.Cuts(icut);			
			
// check the event, if it satisfies cuts or not
            if(icheck == 1){
			
              Ncutevents++;			
		      ev_.AddToFile(outformat,xfile);

// fill the file with the variable
// open the file
             f.open(oxfile, std::ofstream::out | std::ofstream::app);
             if(!f.is_open()) {
                std::cout << "Can not add to an *.exdiff_<var>_<cut> file!\n";   
             }	                   
	
 
 // calc vartofile = ...
              if(ivar == 1){
				  if(IDprocess == 0){vartofile = kin_._t();}
                  else{vartofile = kin_._t_1();}				  
			  } 
               if(ivar == 2){
				  if(IDprocess == 0){vartofile = 0.0;}
                  else{vartofile = kin_._M_c();}				  
			  }
               if(ivar == 3){
				  if(IDprocess == 0){vartofile = kin_._phi();}
                  else{vartofile = kin_._phi_12();}				  
			  }	
               if(ivar == 4){
				  if(IDprocess == 0){vartofile = kin_._y();}
                  else{vartofile = kin_._y_c();}				  
			  }			  
 
 
 // writes variable to the file
              f << std::setprecision(19) << std::setw(34) << std::scientific << vartofile << std::endl;  
// close the file
              f.close();	

			  
		    }

		
	}
  
   std::cout << "GenerateFileCuts is finishing ++++++++++++++++++++++++++++++++++" << std::endl;
	std::cout << "===================================================" <<std::endl;	
    printf("TOTAL C.-S. = %.6f mb \n", CSTOT*Ncutevents/Nevents);   
   

// End BLOCK of output	

    return;	
	
};