/*
 * Particle.cpp
 *
 *  Created on: 17 NOV 2016
 *      Author: Roman Ryutin
 */

#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <stdio.h>
#include <cassert>

#include "../inc/I.h"
#include "../inc/PI.h"
#include "../inc/Particle.h"


// print information
void ExDiff::Particle::Print(){
	std::cout<< " ============================================= " <<std::endl;
	std::cout<< "Particle information" <<std::endl;	
    std::cout<< "ID = " << ID << " ,  Spin*2 + 1 = " << DSpin << " ,  Colour (Anticolour) = " << Colour << "(" << Anticolour << ")" <<std::endl;
    printf("m = %.6f , p*p = %.6f\n", m, pp);	
    printf("p = { %.6f , %.6f , %.6f , %.6f }\n", p[0], p[1], p[2], p[3]);		
//	std::cout<< "m = " << m << " ,  p*p = " << pp << " ,  p = { " << p[0] << " , " << p[1] << " , " << p[2] << " , " << p[3] << " }" <<std::endl;
//    std::cout<< "pT = " << pT << " , p3D = " << p3D << " , y = " << Rapidity(p) << " ,  eta = " << Pseudorapidity(p) <<std::endl;	
	printf("pT = %.6f , p3D = %.6f, y = %.6f, eta = %.6f\n", pT, p3D, Rapidity(p), Pseudorapidity(p));
//	std::cout<< "theta = " << Theta(p) << " , phi = " << Phi(p) <<std::endl;
    printf("theta = %.6f , phi = %.6f\n", Theta(p), Phi(p));
    std::cout<< "on-shell status = " << OnShell() <<std::endl;		
	std::cout<< " ============================================= " <<std::endl;
	return;
};

double ExDiff::Particle::Rapidity(std::vector<double>& p_){
	double y;
	if( std::abs(p_[0] - p_[3]) > 0.000001){
	 y = 0.5*log((p_[0]+p_[3])/(p_[0]-p_[3]));
	} else {
	 y = 1000.0; 
	}
	return y;
};

double ExDiff::Particle::Pseudorapidity(std::vector<double>& p_){
	double eta;	
	if(Theta(p_) > 0.0 && Theta(p_) < PI){	
      eta = - log(tan(0.5*Theta(p_)));
	} else{
      eta = cos(Theta(p_))*1000.0; 	  
	}
    return eta; 	
};

double ExDiff::Particle::Theta(std::vector<double>& p_){
	double theta;
    if(p_[1]*p_[1]+p_[2]*p_[2]+p_[3]*p_[3] > 0.000001){
	 theta = acos(p_[3]/sqrt(p_[1]*p_[1]+p_[2]*p_[2]+p_[3]*p_[3]));
	} else{		
	  if(p_[3]>=0.0){	
       theta = 0.0;
	  } else{
	   theta = PI;  
	  }      	   
	}	
    return theta; 	
};

double ExDiff::Particle::Phi(std::vector<double>& p_){
	double phi;
	if(p_[1]*p_[1]+p_[2]*p_[2] > 0.000001){
     phi = acos(p_[1]/sqrt(p_[1]*p_[1]+p_[2]*p_[2]));
	 if(p_[2] < 0){ phi = 2*PI - acos(p_[1]/sqrt(p_[1]*p_[1]+p_[2]*p_[2]));}
    } else {
	 phi = 0.0;	
	}
    return phi; 	
};

double ExDiff::Particle::Lmult(std::vector<double>& p1_, std::vector<double>& p2_){
    double res = p1_[0]*p2_[0]-p1_[1]*p2_[1]-p1_[2]*p2_[2]-p1_[3]*p2_[3];
    return res; 	
};

std::vector<double> ExDiff::Particle::Lplus(std::vector<double>& p1_, std::vector<double>& p2_){
    std::vector<double> res;
	res.clear();
	for(std::size_t i = 0; i < p1_.size(); i++){
		res.push_back(p1_[i]+p2_[i]);
	}
    return res; 	
};

std::vector<double> ExDiff::Particle::Lminus(std::vector<double>& p1_, std::vector<double>& p2_){
    std::vector<double> res;
	res.clear();
	for(std::size_t i = 0; i < p1_.size(); i++){
		res.push_back(p1_[i]-p2_[i]);
	}
    return res; 	
};

void ExDiff::Particle::Lprint(std::vector<double>& vec_){		  	  
	  std::cout << " { ";
          for (std::size_t j = 0; j < 4; j++) {
             std::cout << vec_[j];
			 if(j < 3) {
				 std::cout <<  " , ";
			 }
          }	  
	  std::cout << " } " << std::endl;	  		
      return;
   };   


int ExDiff::Particle::OnShell(){
	int n = 0;
	double eps = 0.000001;
	double pp_ = pp;
	double mm_ = m*m;
	if(m > eps){
	if(pp_ - mm_ > eps){n = 2;}
	if(pp_ + mm_ < -eps){n = -2;}
	if(pp_ <= -eps && pp_ + mm_ >= eps){n = -1;}
	if(pp_ >= eps && pp_ - mm_ < -eps){n = 1;}
	if(pp_ - mm_ <= eps && pp_ - mm_ >= -eps){n = 0;}
	} else {
		if( m <= eps && m >= -eps){n = 3;}
		if(m < -eps){n = 4;}
	}	
	return n;
};